package uk.ac.leeds.ccg.grids;
/**
 * -------------------------------------------------------------------------
 * This file is distributed as part of the uk.ac.leeds.ccg.grids library.
 * This library is a resource for manipulating spatial data stored as grids.
 * Copyright (C) 2002  Andy Turner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 */
import uk.ac.leeds.ccg.geotools.GeoRectangle;
import uk.ac.leeds.ccg.raster.Raster;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.StreamTokenizer;
import java.io.BufferedReader;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Arrays;
/**
 * ---------------------------
 * Last updated on 25 Nov 2002
 * @author andyt
 * @version 1.0
 * ---------------------------
 * Description:
 * This class controls the manipulation of 2D square cell grids stored in filespace
 */
public class Grid2DSquareCellDoubleFile extends AbstractGrid2DSquareCellDouble {
    
    /**
     * ------
     * Fields
     * ------
     */
    
    /**
     * The directory in which the data is stored
     */
    private String dataDirectory;
    
    /**
     * The file containing all the grid values in row major order
     */
    private File gridFile;
    
    /**
     * The RandomAccessFile for accessing gridFile
     */
    private RandomAccessFile gridRandomAccessFile;
    
    /**
     * -------------------
     * Constructor methods
     * -------------------
     */
    
    /**
     * Creates new Grid2DSquareCellDoubleFile grid with each cell value equal to noDataValue
     * @param gridFile - the file that will contain the grid
     * @param nrows - The number of rows of cells in the constructed grid
     * @param ncols - The number of columns of cells in the constructed grid
     * @param xllcorner - The smallest (left-most) x-coordinate of the grid
     * @param yllcorner - The smallest (bottom-most) x-coordinate of the grid
     * @param cellsize - The cellsize of the grid
     * @param noDataValue - The noDataValue of the grid
     * @param gridStatisticsType - controls the type of AbstractGridStatistics that accompany the grid
     */
    protected Grid2DSquareCellDoubleFile( File gridFile, int nrows, int ncols, double xllcorner, double yllcorner, double cellsize, double noDataValue, int gridStatisticsType ) {
        //System.out.println( "Grid2DSquareCellDoubleFile( File(" + gridFile + "), nrows(" + nrows + "), ncols(" + ncols + "), xllcorner(" + xllcorner + "), yllcorner(" + yllcorner + "), cellsize(" + cellsize + "), noDataValue(" + noDataValue + "), gridStatisticsType(" + gridStatisticsType + ") )" );
        setDataDirectory( gridFile.getPath() );
        setFile( gridFile );
        initGridStatistics( gridStatisticsType );
        RandomAccessFile raf = null;
        try {
            raf = new RandomAccessFile( gridFile, "rw" );
            setRandomAccessFile( raf );
            initNrows( nrows );
            initNcols( ncols );
            setXllcorner( xllcorner );
            setYllcorner( yllcorner );
            setCellsize( cellsize );
            initNoDataValue( noDataValue );
            // Write out values
            double noData = 0.0d;
            if ( noDataValue == noData ) { // There is a fast way to initialise the file!
                raf.seek( ( ( nrows * ncols ) - 1 ) * 8 );
                raf.writeDouble( noData );
            } else {
                raf.seek( 0 );
                for ( int i = 0; i < nrows * ncols; i ++ ) {
                    // Simply populate with noDataValues as stats do not change.
                    raf.writeDouble( noDataValue );
                }
            }
            appendGridStatistics();
        } catch ( java.io.IOException e ) {
            System.out.println( e.toString() + "in Grid2DSquareCellDoubleFile( File(" + gridFile + "), nrows(" + nrows + "), ncols(" + ncols + "), xllcorner(" + xllcorner + "), yllcorner(" + yllcorner + "), cellsize(" + cellsize + "), noDataValue(" + noDataValue + "), gridStatisticsType(" + gridStatisticsType + ") )" );
            System.exit( 0 );
        }
        //System.out.println( toString() );
    }
    
    /*
     * Constructs grid from another grid.
     * @param gridFile - the file that will contain the grid
     * @param inputGrid - the grid from which this will be constructed
     * @param startRowIndex - the inputGrid row index which is the top most row of this grid
     * @param startColIndex - the inputGrid column index which is the left most column of this grid
     * @param endRowIndex - the inputGrid row index which is the bottom most row of this grid
     * @param endColIndex - the inputGrid column index which is the right most column of this grid
     * @param gridStatisticsType - controls the type of AbstractGridStatistics that accompany the grid
     */
    protected Grid2DSquareCellDoubleFile( File gridFile, AbstractGrid2DSquareCellDouble inputGrid, int startRowIndex, int startColIndex, int endRowIndex, int endColIndex, int gridStatisticsType ) {
        //System.out.println( "Grid2DSquareCellDouble( gridFile(" + gridFile + "), AbstractGrid2DSquareCellDouble(" + inputGrid.toString() + "), startRowIndex(" + startRowIndex + "), startColIndex(" + startColIndex + "), endRowIndex(" + endRowIndex + "), endColIndex(" + endColIndex + "), gridStatisticsType(" + gridStatisticsType + ") )" );
        int inputGridNrows = inputGrid.getNrows();
        int inputGridNcols = inputGrid.getNcols();
        double cellsize = inputGrid.getCellsize();
        int nrows = ( endRowIndex - startRowIndex + 1 );
        int ncols = ( endColIndex - startColIndex + 1 );
        initNrows( nrows );
        initNcols( ncols );
        setXllcorner( inputGrid.getXllcorner() + ( double ) startColIndex * cellsize );
        setYllcorner( inputGrid.getYllcorner() + ( double ) ( endRowIndex - inputGridNrows - 1 ) * cellsize );
        setCellsize( cellsize );
        double noDataValue = inputGrid.getNoDataValue();
        initNoDataValue( noDataValue );
        AbstractGridStatistics inputGridStats = inputGrid.getGridStatistics();
        AbstractGridStatistics stats = initGridStatistics( gridStatisticsType );
        double inputGridSparseness = inputGridStats.getSparseness();
        setFile( gridFile );
        try {
            setRandomAccessFile( new RandomAccessFile( gridFile, "rw" ) );
            if ( inputGrid instanceof Grid2DSquareCellDoubleFile ) {
                Grid2DSquareCellDoubleFile inGrid = ( Grid2DSquareCellDoubleFile ) inputGrid;
                File inputGridFile = inGrid.getFile();
                if ( startRowIndex == 0 && startColIndex == 0 && endRowIndex == ( inputGridNrows - 1 ) && endColIndex == ( inputGridNcols - 1 ) ) {
                    // Copy inputGridFile into gridFile
                    // This is an optimisation for SunOS so the file is copied directly by the operating system.
                    // NB. For other platforms similar optimisations can be inserted here.
                    String os = new String( System.getProperty( "os.name" ) );
                    if ( os.equalsIgnoreCase( "SunOS" ) ) {
                        Runtime runtime = java.lang.Runtime.getRuntime();
                        runtime.exec( "cp " + inputGridFile + " " + gridFile.toString() );
                        // Failed optimsiation for Windows XP OS
                        //} else if ( os.equalsIgnoreCase( "Windows XP" ) ) {
                        //    Runtime runtime = java.lang.Runtime.getRuntime();
                        //    //runtime.exec( "copy d:/" + inputGridFile.toString() + " /B d:/" + gridFile.toString() + " /B " );
                    } else {
                        //System.out.println( os );
                        BufferedInputStream bis = new BufferedInputStream( new FileInputStream( inputGridFile ) );
                        BufferedOutputStream bos = new BufferedOutputStream( new FileOutputStream( gridFile ) );
                        for ( int i = 0; i < inputGridFile.length(); i ++ ) {
                            bos.write( bis.read() );
                        }
                        bos.flush();
                        bos.close();
                        bis.close();
                    }
                    stats.updateGridStatisticsFrom( inputGridStats );
                } else {
                    Grid2DSquareCellDoubleFile dummy = new Grid2DSquareCellDoubleFile( gridFile, inputGridFile, startRowIndex, startColIndex, endRowIndex, endColIndex, gridStatisticsType );
                    initNcols( dummy.getNcols() );
                    initNrows( dummy.getNrows() );
                    setXllcorner( dummy.getXllcorner() );
                    setYllcorner( dummy.getYllcorner() );
                    setCellsize( dummy.getCellsize() );
                    initNoDataValue( dummy.getNoDataValue() );
                    stats.updateGridStatisticsFrom( dummy.getGridStatistics() );
                    setDataDirectory( gridFile.getPath() );
                    setFile( dummy.getFile() );
                    setRandomAccessFile( dummy.getRandomAccessFile() );
                    //double sparseness = ((inputGridSparseness * inputGridNrows * inputGridNcols) + ((nrows * ncols) - (inputGridNrows * inputGridNcols))) / (nrows * ncols);
                    //stats.setSparseness(sparseness);
                }
            } else {
                double d1;
                for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                    for ( int j = startColIndex; i <= endColIndex; i ++ ) {
                        d1 = inputGrid.getCell( i, j );
                        if ( d1 != noDataValue ) {
                            initCell( ( i - startRowIndex ), ( j - startColIndex ), d1 );
                        }
                    }
                }
            }
            appendGridStatistics();
        } catch ( java.io.IOException e ) {
            System.out.println( e.toString() + " in Grid2DSquareCellDouble( gridFile(" + gridFile + "), AbstractGrid2DSquareCellDouble(" + inputGrid.toString() + "), startRowIndex(" + startRowIndex + "), startColIndex(" + startColIndex + "), endRowIndex(" + endRowIndex + "), endColIndex(" + endColIndex + "), gridStatisticsType(" + gridStatisticsType + ") )" );
            System.exit( 0 );
        }
        //System.out.println(toString());
    }
    
    /**
     * Constructs from a binary file representation of a grid in the following format:
     *  - The values are written in row major order as double values
     *  - The constructed grid has; ( endColIndex - startColIndex + 1 ) columns, and
     *    ( endRowIndex - startRowIndex + 1 ) rows.
     * @param gridFile - the binary file representaion of the grid being constructed
     * @param inputGridFile - the binary file representaion of the grid to construct from
     * If ( gridFile == inputGridFile ) then the inputGridFile may be overwritten.
     */
    protected Grid2DSquareCellDoubleFile( File gridFile, File inputGridFile, int startRowIndex, int startColIndex, int endRowIndex, int endColIndex, int gridStatisticsType ) {
        //System.out.println( "Grid2DSquareCellDoubleFile( gridFile(" + gridFile + "), inputGridFile(" + inputGridFile + "), startRowIndex(" + startRowIndex + "), startColIndex(" + startColIndex + "), endRowIndex(" + endRowIndex + "), endColIndex(" + endColIndex + "), gridStatisticsType(" + gridStatisticsType + ") )");
        setDataDirectory( gridFile.getPath() );
        setFile( gridFile );
        RandomAccessFile raf = null;
        try {
            raf = new RandomAccessFile( gridFile, "rw" );
        } catch ( java.io.FileNotFoundException e ) {
            System.out.println( e.toString() + " in Grid2DSquareCellDouble( gridFile( " + gridFile + " ), File inputGridFile( " + inputGridFile + " ), startRowIndex( " + startRowIndex + " ), startColIndex( " + startColIndex + " ), endRowIndex( " + endRowIndex + " ), endColIndex( " + endColIndex + " ), gridStatisticsType( " + gridStatisticsType + " ) )" );
            System.exit( 0 );
        }
        setRandomAccessFile( raf );
        initGridStatistics( gridStatisticsType );
        initNrows( endRowIndex - startRowIndex + 1 );
        initNcols( endColIndex - startColIndex + 1 );
        boolean overwrite = false;
        if ( gridFile == inputGridFile ) {
            overwrite = true;
        }
        setDataDirectory( inputGridFile.getPath() );
        String filename = inputGridFile.getName();
        if ( filename.endsWith( "asc" ) ) {
            try {
                StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( inputGridFile ) ) ) );
                st.wordChars( '_', '_' );
                st.parseNumbers();
                st.eolIsSignificant( false );
                st.lowerCaseMode( true );
                int type = st.nextToken();
                st.nextToken();
                int inputGridNcols = ( int )st.nval;
                st.nextToken();
                st.nextToken();
                int inputGridNrows = ( int )st.nval;
                st.nextToken();
                boolean b1 = true;
                if ( st.sval.equalsIgnoreCase( "xllcenter" ) || st.sval.equalsIgnoreCase( "xllcentre" ) ) {
                    b1 = false;
                }
                st.nextToken();
                double xllcorner = st.nval;
                st.nextToken();
                boolean b2 = true;
                if ( st.sval.equalsIgnoreCase( "yllcenter" ) || st.sval.equalsIgnoreCase( "yllcentre" ) ) {
                    b2 = false;
                }
                st.nextToken();
                double yllcorner = st.nval;
                st.nextToken();
                st.nextToken();
                cellsize = st.nval;
                if ( !b1 ) {
                    xllcorner -= ( cellsize / 2.0d );
                }
                if ( !b2 ) {
                    yllcorner -= ( cellsize / 2.0d );
                }
                setXllcorner( xllcorner + ( double ) startColIndex * cellsize );
                setYllcorner( yllcorner + ( ( inputGridNrows -  endRowIndex - 1 ) * cellsize ) );
                type = st.nextToken();
                double noDataValue = Double.NaN;
                if( type == StreamTokenizer.TT_NUMBER ) {
                    // Warning! - if there is no noDataValue and grid values are non-numeric this will break.
                    st.pushBack();
                } else {
                    type = st.nextToken();
                    noDataValue = st.nval;
                }
                initNoDataValue( noDataValue );
                // Having got passed the header st is set so that it recognises E and D as ordinary characters.
                // This is to cope with exponent numbers of the form 0.1E-5 or 0.1D-5; where, E or D means
                // multiplied by ten raised to the power of...
                st.ordinaryChar( 'E' );
                st.ordinaryChar( 'D' );
                // Initialise cell values.
                double d1 = getNoDataValue();
                for ( int i = 0; i < nrows; i ++ ) {
                    for ( int j = 0; j < ncols; j ++ ) {
                        st.nextToken();
                        if ( i >= startRowIndex && i <= endRowIndex ) {
                            if ( j >= startColIndex && j <= endColIndex ) {
                                d1 = st.nval;
                                type = st.nextToken();
                                if ( type != StreamTokenizer.TT_NUMBER && type != StreamTokenizer.TT_EOF ) {
                                    // Either encountered an exponent term number or end of file marker or grid value is non-numeric or something else.
                                    st.nextToken();
                                    d1 = d1 * Math.pow( 10.0, st.nval );
                                } else {
                                    st.pushBack();
                                }
                                initCell( i - startRowIndex, j - startColIndex, d1 );
                            }
                        }
                    }
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + " in Grid2DSquareCellDouble( gridFile( " + gridFile + " ), File inputGridFile( " + inputGridFile + " ), startRowIndex( " + startRowIndex + " ), startColIndex( " + startColIndex + " ), endRowIndex( " + endRowIndex + " ), endColIndex( " + endColIndex + " ), gridStatisticsType( " + gridStatisticsType + " ) )" );
                System.exit( 0 );
            }
        }
        if ( filename.endsWith( "grd" ) ) {
            try {
                RandomAccessFile inputGridRaf = new RandomAccessFile( inputGridFile, "r" );
                int numberOfBytes = ( int ) inputGridRaf.length();
                inputGridRaf.seek( numberOfBytes - ( 13 * 8 ) );
                int inputGridNrows = ( int ) inputGridRaf.readDouble();
                int inputGridNcols = ( int ) inputGridRaf.readDouble();
                int nrows = endRowIndex - startRowIndex + 1;
                int ncols = endColIndex - startColIndex + 1;
                double inputGridXllcorner = inputGridRaf.readDouble();
                double inputGridYllcorner = inputGridRaf.readDouble();
                double cellsize = inputGridRaf.readDouble();
                double noDataValue = inputGridRaf.readDouble();
                initNrows( nrows );
                initNcols( ncols );
                setXllcorner( inputGridXllcorner + ( ( ( double ) startColIndex ) * cellsize ) );
                setYllcorner( inputGridYllcorner + ( ( ( double ) ( endRowIndex - inputGridNrows + 1 ) ) * cellsize ) );
                setCellsize( cellsize );
                setNoDataValue( noDataValue );
                // If ( startRowIndex == 0 && startColIndex == 0 && endRowIndex == ( inputGridNrows - 1 ) && endColIndex == ( inputGridNcols - 1 ) )
                // Then there is a fast way to construct:
                if ( startRowIndex == 0 && startColIndex == 0 && endRowIndex == ( inputGridNrows - 1 ) && endColIndex == ( inputGridNcols - 1 ) ) {
                    if ( ! overwrite ) {
                        // Copy inputGridFile into gridFile
                        String os = new String( System.getProperty( "os.name" ) );
                        //System.out.println( os );
                        // Optimisation for SunOS.
                        if ( os.equalsIgnoreCase( "SunOS" ) ) {
                            Runtime runtime = java.lang.Runtime.getRuntime();
                            runtime.exec( "cp " + inputGridFile.toString() + " " + gridFile.toString() );
                            // Failed optimsiation for Windows XP OS
                            //if ( os.equalsIgnoreCase( "Windows XP" ) ) {
                            //    Runtime runtime = java.lang.Runtime.getRuntime();
                            //    //runtime.exec( "C:/WINDOWS/System32/cmd.exe copy d:/" + inputGridFile.toString() + " /B d:/" + gridFile.toString() + " /B " );
                        } else {
                            BufferedInputStream bis = new BufferedInputStream( new FileInputStream( inputGridFile ) );
                            BufferedOutputStream bos = new BufferedOutputStream( new FileOutputStream( gridFile ) );
                            for ( int i = 0; i < numberOfBytes; i ++ ) {
                                bos.write( bis.read() );
                            }
                            bos.flush();
                            bos.close();
                            bis.close();
                        }
                        setRandomAccessFile( new RandomAccessFile( gridFile, "rw" ) );
                    } else {
                        setRandomAccessFile( inputGridRaf );
                    }
                } else {
                    if ( overwrite ) {
                        Utilities u = new Utilities();
                        gridFile = u.makeTemporaryFile( getDataDirectory() );
                    }
                    double noData = 0.0d;
                    if ( noDataValue == noData ) { // There is a fast way to initialise the file!
                        raf.seek( ( ( nrows * ncols ) - 1 ) * 8 );
                        raf.writeDouble( noData );
                    } else {
                        raf.seek( 0 );
                        for ( int i = 0; i < nrows * ncols; i ++ ) {
                            raf.writeDouble( noDataValue );
                        }
                    }
                    // Replace coincident values from input
                    if ( !( startColIndex >= inputGridNcols || endColIndex < 0 || startRowIndex >= inputGridNrows || endRowIndex < 0 ) ) {
                        // Only need do something if there are coincident values.
                        int caseType=0;
                        // There are 16 cases to deal with.
                        if ( startColIndex <= 0 && startRowIndex <= 0 && endRowIndex >= inputGridNrows - 1 && endColIndex >= inputGridNcols -1 ) {caseType = 1; }
                        if ( startColIndex <= 0 && startRowIndex <= 0 && endRowIndex >= inputGridNrows - 1 && endColIndex < inputGridNcols -1 ) {caseType = 2; }
                        if ( startColIndex <= 0 && startRowIndex <= 0 && endRowIndex < inputGridNrows - 1 && endColIndex >= inputGridNcols -1 ) {caseType = 3; }
                        if ( startColIndex <= 0 && startRowIndex <= 0 && endRowIndex < inputGridNrows - 1 && endColIndex < inputGridNcols -1 ) {caseType = 4; }
                        if ( startColIndex > 0 && startRowIndex <= 0 && endRowIndex >= inputGridNrows - 1 && endColIndex >= inputGridNcols -1 ) {caseType = 5; }
                        if ( startColIndex > 0 && startRowIndex <= 0 && endRowIndex >= inputGridNrows - 1 && endColIndex < inputGridNcols -1 ) {caseType = 6; }
                        if ( startColIndex > 0 && startRowIndex <= 0 && endRowIndex < inputGridNrows - 1 && endColIndex >= inputGridNcols -1 ) {caseType = 7; }
                        if ( startColIndex > 0 && startRowIndex <= 0 && endRowIndex < inputGridNrows - 1 && endColIndex < inputGridNcols -1 ) {caseType = 8; }
                        if ( startColIndex <= 0 && startRowIndex > 0 && endRowIndex >= inputGridNrows - 1 && endColIndex >= inputGridNcols -1 ) {caseType = 9; }
                        if ( startColIndex <= 0 && startRowIndex > 0 && endRowIndex >= inputGridNrows - 1 && endColIndex < inputGridNcols -1 ) {caseType = 10; }
                        if ( startColIndex <= 0 && startRowIndex > 0 && endRowIndex < inputGridNrows - 1 && endColIndex >= inputGridNcols -1 ) {caseType = 11; }
                        if ( startColIndex <= 0 && startRowIndex > 0 && endRowIndex < inputGridNrows - 1 && endColIndex < inputGridNcols -1 ) {caseType = 12; }
                        if ( startColIndex > 0 && startRowIndex > 0 && endRowIndex >= inputGridNrows - 1 && endColIndex >= inputGridNcols -1 ) {caseType = 13; }
                        if ( startColIndex > 0 && startRowIndex > 0 && endRowIndex >= inputGridNrows - 1 && endColIndex < inputGridNcols -1 ) {caseType = 14; }
                        if ( startColIndex > 0 && startRowIndex > 0 && endRowIndex < inputGridNrows - 1 && endColIndex >= inputGridNcols -1 ) {caseType = 15; }
                        if ( startColIndex > 0 && startRowIndex > 0 && endRowIndex < inputGridNrows - 1 && endColIndex < inputGridNcols - 1) {caseType = 16; }
                        //System.out.print( "caseType " + caseType );
                        switch (caseType) {
                            case 1 :
                                inputGridRaf.seek(0);
                                for ( int i = 0; i < inputGridNrows; i ++ ) {
                                    for ( int j = 0; j <inputGridNcols; j ++ ) {
                                        initCell( i - startRowIndex, j - startColIndex, inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 2 :
                                for ( int i = 0; i < inputGridNrows; i ++ ) {
                                    inputGridRaf.seek( ( i * inputGridNcols) * 8 );
                                    for ( int j = 0; j <= endColIndex; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 3 :
                                for ( int i = 0; i <= endRowIndex; i ++ ) {
                                    inputGridRaf.seek( ( i * inputGridNcols ) * 8 );
                                    for ( int j = 0; j < inputGridNcols; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 4 :
                                for ( int i = 0; i <= endRowIndex; i ++ ) {
                                    inputGridRaf.seek( ( i * inputGridNcols ) * 8 );
                                    for ( int j = 0; j <= endColIndex; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 5 :
                                for ( int i = 0; i < inputGridNrows; i ++ ) {
                                    inputGridRaf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                    for ( int j = startColIndex; j < inputGridNcols; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 6 :
                                for ( int i = 0; i < inputGridNrows; i ++ ) {
                                    inputGridRaf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                    for ( int j = startColIndex; j <= endColIndex; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 7 :
                                for ( int i = 0; i <=endRowIndex; i ++ ) {
                                    inputGridRaf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                    for ( int j = startColIndex; j < inputGridNcols; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 8 :
                                for ( int i = 0; i <= endRowIndex; i ++ ) {
                                    inputGridRaf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                    for ( int j = 0; j <= endColIndex; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 9 :
                                for ( int i = startRowIndex; i < inputGridNrows; i ++ ) {
                                    inputGridRaf.seek( ( i * inputGridNcols ) * 8 );
                                    for ( int j = 0; j < inputGridNcols; j ++ ) {
                                        initCell( ( i - startRowIndex), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 10 :
                                for ( int i = startRowIndex; i < inputGridNrows; i ++ ) {
                                    inputGridRaf.seek( ( i * inputGridNcols ) * 8 );
                                    for ( int j = 0; j <= endColIndex; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 11 :
                                for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                                    inputGridRaf.seek( ( i * inputGridNcols ) * 8 );
                                    for ( int j = 0; j < inputGridNcols; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                            case 12 :
                                for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                                    inputGridRaf.seek( ( i * inputGridNcols ) * 8 );
                                    for ( int j = 0; j <= endColIndex; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 13 :
                                for ( int i = startRowIndex; i < inputGridNrows; i ++ ) {
                                    inputGridRaf.seek( ( ( i * inputGridNcols) + startColIndex ) * 8 );
                                    for ( int j = startColIndex; j < inputGridNcols; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 14 :
                                for ( int i = startRowIndex; i < inputGridNrows; i ++ ) {
                                    inputGridRaf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                    for ( int j = startColIndex; j <= endColIndex; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 15 :
                                for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                                    inputGridRaf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                    for ( int j = startColIndex; j < inputGridNcols; j ++) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                            case 16 :
                                for ( int i = startRowIndex; i <= endRowIndex; i ++ ) {
                                    inputGridRaf.seek( ( ( i * inputGridNcols ) + startColIndex ) * 8 );
                                    for ( int j = startColIndex; j <= endColIndex; j ++ ) {
                                        initCell( ( i - startRowIndex ), ( j - startColIndex ), inputGridRaf.readDouble() );
                                    }
                                }
                                break;
                        }
                    }
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + " in Grid2DSquareCellDouble( gridFile( " + gridFile + " ), File inputGridFile( " + inputGridFile + " ), startRowIndex( " + startRowIndex + " ), startColIndex( " + startColIndex + " ), endRowIndex( " + endRowIndex + " ), endColIndex( " + endColIndex + " ), gridStatisticsType( " + gridStatisticsType + " ) )" );
                System.exit( 0 );
            }
        }
        if ( overwrite ) {
            try {
                // Copy gridFile into inputGridFile
                String os = new String( System.getProperty( "os.name" ) );
                //System.out.println( os );
                // Optimisation for SunOS.
                if ( os.equalsIgnoreCase( "SunOS" ) ) {
                    Runtime runtime = java.lang.Runtime.getRuntime();
                    runtime.exec( "cp " + gridFile.toString() + " " + inputGridFile.toString() );
                    // Failed optimsiation for Windows XP OS
                    //if ( os.equalsIgnoreCase( "Windows XP" ) ) {
                    //    Runtime runtime = java.lang.Runtime.getRuntime();
                    //    //runtime.exec( "copy " + gridFile.toString() + " " + inputGridFile.toString() );
                } else {
                    BufferedInputStream bis = new BufferedInputStream( new FileInputStream( gridFile ) );
                    BufferedOutputStream bos = new BufferedOutputStream( new FileOutputStream( inputGridFile ) );
                    for ( int i = 0; i < raf.length(); i ++ ) {
                        bos.write( bis.read() );
                    }
                    bos.flush();
                    bos.close();
                    bis.close();
                }
                raf.close();
                gridFile.delete();
                setRandomAccessFile( new RandomAccessFile( inputGridFile, "rw" ) );
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + " in Grid2DSquareCellDouble( gridFile( " + gridFile + " ), File inputGridFile( " + inputGridFile + " ), startRowIndex( " + startRowIndex + " ), startColIndex( " + startColIndex + " ), endRowIndex( " + endRowIndex + " ), endColIndex( " + endColIndex + " ), gridStatisticsType( " + gridStatisticsType + " ) )" );
                System.exit( 0 );
            }
        }
        try {
            appendGridStatistics();
        } catch ( java.io.IOException e ) {
            System.out.println( e.toString() + " in Grid2DSquareCellDouble( gridFile( " + gridFile + " ), File inputGridFile( " + inputGridFile + " ), startRowIndex( " + startRowIndex + " ), startColIndex( " + startColIndex + " ), endRowIndex( " + endRowIndex + " ), endColIndex( " + endColIndex + " ), gridStatisticsType( " + gridStatisticsType + " ) )" );
            System.exit( 0 );
        }
        //System.out.println( toString() );
    }
    
    /**
     * Constructs from a uk.ac.leeds.ccg.raster.Raster
     * Grid2DSquareCellDoubleFile( gridFile, raster, gridStatisticsType );
     * @param gridFile - the file for storing the constructed grid data
     * @param raster - the uk.ac.leeds.ccg.raster.Raster to construct from
     */
    protected Grid2DSquareCellDoubleFile( File gridFile, Raster raster, int gridStatisticsType ) {
        //System.out.println( "Grid2DSquareCellDoubleFile( gridFile(" + gridFile + "), Raster(" + raster.toString() + "), gridStatisticsType(" + gridStatisticsType + ") )" );
        initGridStatistics( gridStatisticsType );
        int ncols = raster.getWidth();
        int nrows = raster.getHeight();
        initNrows( nrows );
        initNcols( ncols );
        setXllcorner( raster.getOriginx() );
        setYllcorner( raster.getOriginy() );
        setCellsize( raster.getCellSize() );
        initNoDataValue( raster.getMissingValueCode() );
        //initNoDataValue(raster.getNoDataValue());
        setDataDirectory( gridFile.getPath() );
        setFile( gridFile );
        try {
            setRandomAccessFile( new RandomAccessFile( gridFile, "rw" ) );
            for ( int i = 0; i < nrows; i ++ ) {
                for ( int j = 0; j < ncols; j ++ ) {
                    initCell( i, j, raster.getCell( i, j ) );
                }
            }
            appendGridStatistics();
        } catch ( java.io.IOException e ) {
            System.out.println( e.toString() + " in Grid2DSquareCellDoubleFile( gridFile(" + gridFile + "), Raster(" + raster.toString() + "), gridStatisticsType(" + gridStatisticsType + ") )" );
            System.exit( 0 );
        }
        //System.out.println( toString() );
    }
    
    /**
     * ---------------
     * General Methods
     * ---------------
     */
    
    /**
     * Returns a description of the grid.
     */
    public String toString() {
        AbstractGridStatistics stats = getGridStatistics();
        return "File( " + getFile() + " ), ncols( " + getNcols() + " ), nrows( " + getNrows() + " ), xllcorner( " + getXllcorner() + " ), yllcorner( " + getYllcorner() + " ), cellsize( " + getCellsize() + " ), noDataValue( " + getNoDataValue() + " ), " + stats.toString();
    }
    
    /**
     * Returns the data type
     */
    public int getDataType() {
        return 1;
    }
    
    /**
     * Deletes gridFile
     */
    public void clear() {
        try {
            getRandomAccessFile().close();
        } catch ( java.io.IOException e) {
            System.out.println( e + " in clear()" );
        }
        getFile().delete();
    }
    
    /**
     * Returns dataDirectory.
     */
    public String getDataDirectory() {
        return this.dataDirectory;
    }
    
    /**
     * Sets dataDirectory to the String passed in
     */
    protected final void setDataDirectory( String dataDirectory ) {
        this.dataDirectory = dataDirectory;
    }
    
    /**
     * Returns gridFile.
     */
    public final File getFile() {
        return this.gridFile;
    }
    
    /**
     * Sets gridFile.
     */
    protected final void setFile(File gridFile) {
        this.gridFile=gridFile;
    }
    
    /**
     * Returns gridRandomAccessFile.
     */
    public final RandomAccessFile getRandomAccessFile() {
        return gridRandomAccessFile;
    }
    
    /**
     * Sets gridRandomAccessFile.
     */
    private final void setRandomAccessFile( RandomAccessFile raf ) {
        this.gridRandomAccessFile=raf;
    }
    
    /**
     * Returns the value at row i, column j else returns noDataValue.
     */
    public double getCell( int i, int j ) {
        if ( inGrid( i, j ) ) {
            try {
                RandomAccessFile raf = getRandomAccessFile();
                raf.seek( ( ( i * getNcols() ) + j ) * 8 );
                //raf.seek( getCellID( i, j ) * 8 );
                return raf.readDouble();
            } catch ( java.io.IOException e ) {
                if ( e instanceof java.io.FileNotFoundException ) {
                    System.out.println( e.toString() + " in getCell(" + i + "," + j + ")" );
                    System.exit( 0 );
                } // Else trying to seek to a position or read a value that does not exist so return noDataValue!
            }
        }
        return getNoDataValue();
    }
    
    /**
     * Returns the value at row i column j and sets the value at row i column j to d1.
     */
    public double setCell(int i, int j, double d1) {
        //System.out.println( "setCell( row( " + i + " ), col( " + j + " ), value( " + d1 + " ) ) {" );
        double noDataValue = getNoDataValue();
        if ( inGrid( i, j ) ) {
            AbstractGridStatistics stats = getGridStatistics();
            int gridStatisticsType = stats.getGridStatisticsType();
            try {
                RandomAccessFile raf = getRandomAccessFile();
                int nrows = getNrows();
                int ncols = getNcols();
                double cellAsProportionOfGrid = ( 1.0d / ( double ) ( nrows * ncols ) );
                raf.seek( ( ( i * ncols ) + j ) * 8 );
                //raf.seek( getCellID( i, j ) * 8 );
                if ( d1 == noDataValue ) { stats.setSparseness( stats.getSparseness() + cellAsProportionOfGrid ); }
                double d2 = raf.readDouble();
                if ( d2 == noDataValue ) { stats.setSparseness( stats.getSparseness() - cellAsProportionOfGrid ); }
                if ( gridStatisticsType == 0 ) {
                    stats.doMin( d1 );
                    if ( d2 == stats.getMin() ) { stats.setMinCount( stats.getMinCount() - 1 ); }
                    stats.doMax( d1 );
                    if ( d2 == stats.getMax() ) { stats.setMaxCount( stats.getMaxCount() - 1 ); }
                    if ( d1 != noDataValue ) { stats.setSum( stats.getSum() + d1 ); }
                    if ( d2 != noDataValue ) { stats.setSum( stats.getSum() - d2 ); }
                }
                raf.seek( ( ( i * ncols ) + j ) * 8 );
                //raf.seek( getCellID( i, j ) * 8 );
                raf.writeDouble( d1 );
                return d2;
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + " in setCell(" + i + "," + j + "," + d1 + ")" );
                e.printStackTrace();
                System.exit( 0 );
            }
        }
        return noDataValue;
    }
    
    /**
     * Initilises the value at row i column j to d1.
     */
    protected void initCell( int i, int j, double d1 ) {
        //System.out.println( "initCell( row( " + i + " ), col( " + j + " ), value( " + d1 + " ) ) {" );
        if ( inGrid( i, j ) ) {
            AbstractGridStatistics stats = getGridStatistics();
            int gridStatisticsType = stats.getGridStatisticsType();
            try {
                int ncols = getNcols();
                double noDataValue = getNoDataValue();
                double sparseness = stats.getSparseness();
                RandomAccessFile raf = getRandomAccessFile();
                raf.seek( ( ( i * ncols ) + j ) * 8 );
                raf.writeDouble( d1 );
                if ( d1 != noDataValue ) {
                    stats.setSparseness(  sparseness - ( 1.0d / ( double ) ( getNrows() * ncols ) ) );
                    if ( gridStatisticsType == 0 ) {
                        double sum = stats.getSum();
                        stats.doMin( d1 );
                        stats.doMax( d1 );
                        stats.setSum( sum + d1 );
                    }
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() + " in initCell( " + i + ", " + j + ", " + d1 + " )" );
                System.exit( 0 );
            }
        } else {
            System.out.println( "Warning!!! InGrid(" + i + ", " + j + ") == false in initCell( " + i + ", " + j + ", " + d1 + " )" );
        }
    }
    
    protected void appendGridStatistics() throws java.io.IOException {
        //System.out.println( "appendGridStatistics()" );
        RandomAccessFile raf = getRandomAccessFile();
        raf.seek( raf.length() );
        raf.writeDouble( ( double ) getNrows() );
        raf.writeDouble( ( double ) getNcols() );
        raf.writeDouble( getXllcorner() );
        raf.writeDouble( getYllcorner() );
        raf.writeDouble( getCellsize() );
        raf.writeDouble( getNoDataValue() );
        double[] statsToArray = this.getGridStatistics().toArray();
        raf.writeDouble( statsToArray[ 0 ] );
        raf.writeDouble( statsToArray[ 1 ] );
        raf.writeDouble( statsToArray[ 2 ] );
        raf.writeDouble( statsToArray[ 3 ] );
        raf.writeDouble( statsToArray[ 4 ] );
        raf.writeDouble( statsToArray[ 5 ] );
        raf.writeDouble( statsToArray[ 6 ] );
    }
}