package uk.ac.leeds.ccg.grids;
/**
 * -------------------------------------------------------------------------
 * This file is distributed as part of the uk.ac.leeds.ccg.grids library.
 * This library is a resource for manipulating spatial data stored as grids.
 * Copyright (C) 2002  Andy Turner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * -------------------------------------------------------------------------
 */
/**
 * ---------------------------
 * Last updated on 27 Nov 2002
 * @author andyt
 * @version 1.0
 * ---------------------------
 * Description:
 * This class is instatiated in a constructor extended from AbstractsGrid2DSquareCellDouble
 * It does not attempt to keep grid statistics up-to-date as the underlying data is changed.
 *
 * Notes:
 * 1. This extention of AbstractGridStatistics should be instantiated if the min,
 *  max or sum of the grid are not needed or unless a considerable number of cell
 *  values are changed prior to them being needed.
 * 2. Having asked for the min, max or sum then it is inexpensive to convert to
 *  revert to GridStatistics0.
 */
public class GridStatistics1 extends AbstractGridStatistics {
    
    /**
     * ------
     * Fields
     * ------
     */
    
    private boolean isUpToDate;
    
    /**
     * -------------------
     * Constructor methods
     * -------------------
     */
    
    /**
     * Creates a new instance of GridStatistics1
     */
    public GridStatistics1(AbstractGrid2DSquareCellDouble grid) {
        // Initialise
        init( grid );
        setIsUpToDate( false );
    }
    
    /**
     * Creates a new instance of GridStatistics1 to replace gridStatistics
     * @param gridStatistics - the gridStatistics which this will replace
     */
    public GridStatistics1( AbstractGridStatistics gridStatistics ) {
        AbstractGrid2DSquareCellDouble grid = gridStatistics.getGrid();
        int gridStatisticsType = gridStatistics.getGridStatisticsType();
        GridStatistics0 dummy0 = ( GridStatistics0 ) gridStatistics;
        sum = dummy0.sum;
        setSparseness( dummy0.getSparseness() );
        setMax( dummy0.max );
        maxCount = dummy0.maxCount;
        min = dummy0.min;
        minCount = dummy0.minCount;
        setIsUpToDate( dummy0.getIsUpToDate() );
        setGrid( grid );
        grid.setGridStatistics( this );
    }
    
    /**
     * ---------------
     * General methods
     * ---------------
     */
    
    /**
     * Returns the type of gridStatistics:
     * 0 - GridStatistics0
     * 1 - GridStatistics1
     */
    public int getGridStatisticsType() {
        return 1;
    }
    
    /**
     * Returns true iff stats are upto date
     */
    public boolean getIsUpToDate() {
        return this.isUpToDate;
    }
    
    /**
     * Sets isUpToDate to value
     */
    protected void setIsUpToDate( boolean value ) {
        this.isUpToDate = value;
    }
    
    /**
     * Do nothing
     */
    protected void setSum( double value ) {}
    
    /**
     * Returns the minimum value in the Grid.
     */
    public double getMin() {
        if ( getIsUpToDate() ) {
            return this.min;
        } else {
            doStats();
        }
        return this.min;
    }
    
    /**
     * Do nothing
     */
    protected void setMin( double value ) {}
    
    /**
     * Returns a count of the number of cells in the grid with the minimum value
     */
    public int getMinCount() {
        if ( getIsUpToDate() ) {
            return this.minCount;
        } else {
            doStats();
        }
        return this.minCount;
    }
    
    /**
     * Do nothing
     */
    protected void setMinCount( int minCount ) {}
    
    /**
     * Do nothing
     */
    protected void doMin( double d1 ) {}
    
    /**
     * Returns the maximum value in the Grid.
     */
    public double getMax() {
        if ( getIsUpToDate() ) {
            return this.max;
        } else {
            doStats();
        }
        return this.max;
    }
    
    /**
     * Do nothing
     */
    protected void setMax( double value ) {}
    
    /**
     * Do nothing
     */
    protected void setMaxCount( int value ) {}
    
    /**
     * Returns a count of the number of cells in the grid with the maximum value
     */
    public int getMaxCount() {
        if ( getIsUpToDate() ) {
            return this.maxCount;
        } else {
            doStats();
        }
        return this.maxCount;
    }
    
    /**
     * Do nothing
     */
    protected void doMax( double d1 ) {}
    
}
